/* See LICENSE file for copyright and license details. */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "../headers/file.h"
#include "../headers/neural_net.h"

typedef struct {
	File *i_file;
	File *t_file;
	double t_factor;
	int rows_num;
	double *output_deltas;
	double **small_deltas;
	double ***weight_deltas;
} TrainingInfo;

static TrainingInfo *new_training_info(NeuralNet *n, char *i_name,
		char *t_name, double t_factor);
static void free_training_info(NeuralNet *n, TrainingInfo *t);
static void init_inputs(NeuralNet *n);
static void init_weights(NeuralNet *n);
static void set_input_layer(NeuralNet *n, TrainingInfo *t, int i_file_row);
static void calculate_outputs(NeuralNet *n);
static void calculate_output_deltas(NeuralNet *n, TrainingInfo *t, int row);
static void calculate_small_deltas(NeuralNet *n, TrainingInfo *t);
static void calculate_weight_deltas(NeuralNet *n, TrainingInfo *t);
static void calculate_weights(NeuralNet *n, TrainingInfo *t);
static void print_weights(NeuralNet *n);

void
train_by_backpropagation(NeuralNet *n, int runs, double t_factor, char *i_name,
		char *t_name)
{
	int i, j;
	TrainingInfo *t;

	t = new_training_info(n, i_name, t_name, t_factor);
	init_inputs(n);
	init_weights(n);

	for (i=0; i<runs; i++) {
		for (j=0; j<t->rows_num; j++) {
			set_input_layer(n, t, j);
			calculate_outputs(n);
			calculate_output_deltas(n, t, j);
			calculate_small_deltas(n, t);
			calculate_weight_deltas(n, t);
			calculate_weights(n, t);
		}
	}
	print_weights(n);

	free_training_info(n, t);
}

static TrainingInfo
*new_training_info(NeuralNet *n, char *i_name, char *t_name, double t_factor)
{
	int i, j;
	TrainingInfo *t;

	t = malloc(sizeof(TrainingInfo));
	t->i_file = new_file(i_name);
	t->t_file = new_file(t_name);
	t->t_factor = t_factor;
	t->rows_num = get_rows_num(t->i_file);
	t->output_deltas = calloc(n->l[n->l_num-1]->p_num, sizeof(double));
	t->small_deltas  = calloc(n->l_num-1, sizeof(double *));
	for (i=0; i<n->l_num; i++) {
		t->small_deltas[i]  = calloc(n->l[i]->p_num, sizeof(double));
	}
	t->weight_deltas = calloc(n->l_num-1, sizeof(double **));
	for (i=1; i<n->l_num; i++) {
		t->weight_deltas[i] = calloc(n->l[i]->p_num, sizeof(double *));
		for (j=0; j<n->l[i]->p_num; j++) {
			t->weight_deltas[i][j] = calloc(n->l[i-1]->p_num,
				sizeof(double));
		}
	}

	return t;
}

static void
free_training_info(NeuralNet *n, TrainingInfo *t)
{
	int i, j;

	free_file(t->i_file);
	free_file(t->t_file);

	free(t->output_deltas);
	for (i=1; i<n->l_num; i++) {
		free(t->small_deltas[i]);
	}
	free(t->small_deltas);
	for (i=1; i<n->l_num; i++) {
		for (j=0; j<n->l[i]->p_num; j++) {
			free(t->weight_deltas[i][j]);
		}
		free(t->weight_deltas[i]);
	}
	free(t->weight_deltas);
	free(t);
}

static void
init_inputs(NeuralNet *n)
{
	int i;

	for (i=0; i<n->l_num-1; i++) {
		n->l[i]->p[0]->output = 1;
	}
}

static void
init_weights(NeuralNet *n)
{
	int i, j, k;

	srand(time(NULL));
	for (i=1; i<n->l_num; i++) {
		for (j=1; j<n->l[i]->p_num; j++) {
			for (k=0; k<n->l[i]->p[j]->i_num; k++) {
				n->l[i]->p[j]->weights[k] = (double) rand()
					/ RAND_MAX;
			}
		}
	}
}

static void
set_input_layer(NeuralNet *n, TrainingInfo *t, int i_file_row)
{
	int i;

	for (i=1; i<n->l[0]->p_num-1; i++) {
		n->l[0]->p[i]->output =
			get_entry(t->i_file, i_file_row, i, ',');
	}
}

static void
calculate_outputs(NeuralNet *n)
{
	int i, j, k;
	int x;

	for (i=1; i<n->l_num; i++) {
		for (j=1; j<n->l[i]->p_num; j++) {
			x = 0;
			for (k=0; k<n->l[i]->p[j]->i_num; k++) {
				x -= n->l[i]->p[j]->weights[k] *
					n->l[i-1]->p[k]->output;
			}
			n->l[i]->p[j]->output = 1 / (1 + exp(x));
		}
	}
}

static void
calculate_output_deltas(NeuralNet *n, TrainingInfo *t, int row)
{
	int i;

	for (i=0; i<n->l[n->l_num-1]->p_num; i++) {
		t->output_deltas[i] = get_entry(t->t_file, row, i, ',')
			- n->l[n->l_num-1]->p[i]->output;
	}
}

static void
calculate_small_deltas(NeuralNet *n, TrainingInfo *t)
{
	int i, j, k;
	double sum;

	for (i=0; i<n->l[n->l_num-1]->p_num; i++) {
		t->small_deltas[n->l_num-1][i] = t->output_deltas[i]
			* (1 - n->l[n->l_num-1]->p[i]->output)
			* n->l[n->l_num-1]->p[i]->output;
	}
	
	sum = 0;
	for (i=0; i<n->l[n->l_num-2]->p_num; i++) {
		sum += t->small_deltas[n->l_num-1][i];
	}
	for (i=1; i<n->l[n->l_num-2]->p_num; i++) {
		for (j=0; j<n->l[n->l_num-1]->p_num; j++) {
			t->small_deltas[n->l_num-2][i-1]
				= (1 - n->l[n->l_num-2]->p[j]->output)
				* n->l[n->l_num-2]->p[j]->output
				* sum
				* n->l[n->l_num-1]->p[j]->weights[i];
		}
	}

	for (i=n->l_num-3; i>0; i--) {
		sum = 0;
		for (j=1; j<n->l[i+1]->p_num; j++) {
			sum += t->small_deltas[i+1][j-1];
		}
		for (j=0; j<n->l[i]->p_num; j++) {
			for (k=1; k<n->l[i+1]->p_num; k++) {
				t->small_deltas[i][j]
					= (1 - n->l[i]->p[k]->output)
					* n->l[i]->p[k]->output
					* sum
					* n->l[i+1]->p[k]->weights[j];
			}
		}
	}
}

static void
calculate_weight_deltas(NeuralNet *n, TrainingInfo *t)
{
	int i, j, k;

	for (i=n->l_num-1; i>0; i--) {
		for (j=0; j<n->l[i]->p_num; j++) {
			for (k=0; k<n->l[i-1]->p_num; k++) {
				t->weight_deltas[i][j][k] = t->t_factor
					* t->small_deltas[i][j]
					* n->l[i-1]->p[k]->output;
			}
		}
	}
}

static void
calculate_weights(NeuralNet *n, TrainingInfo *t)
{
	int i, j, k;

	for (i=n->l_num-1; i>0; i--) {
		for (j=0; j<n->l[i]->p_num; j++) {
			for (k=0; k<n->l[i-1]->p_num; k++) {
				n->l[i]->p[j]->weights[k]
					= n->l[i]->p[j]->weights[k]
					- t->weight_deltas[i][j][k];
			}
		}
	}
}

static void
print_weights(NeuralNet *n)
{
	int i, j, k;

	for (i=1; i<n->l_num-1; i++) {
		fprintf(stderr, "L %d", i);
		for (j=1; j<n->l[i]->p_num; j++) {
			fprintf(stderr, "\tP %d\n", j);
			for (k=0; k<n->l[i-1]->p_num; k++) {
				printf("%lf,", n->l[i]->p[j]->weights[k]);
			}
			printf("\n");
		}
		printf("\n");
	}
	fprintf(stderr, "L %d", i);
	for (i=0; i<n->l[n->l_num-1]->p_num; i++) {
		fprintf(stderr, "\tP %d\n", i);
		for (j=0; j<n->l[n->l_num-2]->p_num; j++) {
			printf("%lf,", n->l[n->l_num-1]->p[i]->weights[j]);
		}
		printf("\n");
	}
	printf("\n");
}
